"""The ``Loggable`` module provides tools for configuring the logging of Sapicore models.
It supports logging simple properties to tensorboard as well as arbitrary properties,
including tensors and arrays, to HDF5 files.

Individual properties can be included or excluded from logging using a config
dict that can be saved to a YAML, similarly to :mod:`tree-config` configuration.

Loggable
--------

:class:`~sapicore.utils.loggable.Loggable` defines an API for properties to support opting-in to logging.
Users may list loggable properties by including them in :attr:`~utils.loggable.Loggable._loggable_props_`
on a per class basis.

Subclassing extends :attr:`~utils.loggable.Loggable._loggable_props_`, which are accumulated across
all superclasses when :attr:`~utils.loggable.Loggable.loggable_props` is called. The latter also
updates the cache list :attr:`~utils.loggable.Loggable._loggable_props`.
Setting `utils.loggable.Loggable._loggable_props_`, in and of itself, will NOT update the cache.

To support the logging of nested objects, we use :attr:`~utils.loggable.Loggable._loggable_children_`
to list all objects that should be inspected for loggable properties. These are similarly listed in
:attr:`~utils.loggable.Loggable.loggable_children`.

API
---

Users may obtain the loggable properties as YAML files, dictionaries, or flat lists of properties using
the following functions:

:func:`read_loggable_from_object`, :func:`read_loggable_from_file`, :func:`update_loggable_from_object`,
:func:`dump_loggable`, :func:`load_save_get_loggable_properties`, and :func:`get_loggable_properties`.

They all provide a mapping from the property name to a boolean indicting whether the property should be logged.

Example
-------
The following class inherits from Loggable:

    >>> class Example(Loggable):
    >>>     _loggable_props_: tuple[str] = ("voltage",)
    >>>     def __init__(self, voltage: list[float]):
    >>>         super().__init__()
    >>>         self.voltage = voltage

After instantiating an object, refresh the cache and obtain the loggable properties:

    >>> dummy = Example(voltage=[1, 2, 3])
    >>> print(dummy.loggable_props)
    ['voltage']

To check which attributes are loggable:

    >>> read_loggable_from_object(dummy, True)
    {'voltage': True}

Note
----
:func:`~utils.loggable.log_tensor_board` supports taking the flat list of properties and
logging all the properties that are set to be logged to the tensorboard
format on disk using a :class:`~torch.utils.tensorboard.SummaryWriter` instance.

Similarly, :class:`NixLogWriter` supports logging arbitrary tensors and numpy
arrays to a nix HDF5 file. :class:`NixLogReader` can then be used to load the
data from the nix file into scalars or numpy arrays of the original shape.

Warning
-------
This module may change significantly in Sapicore > 0.3.0.

"""
from typing import Tuple, List, Dict, Any, Union, Optional
from pathlib import Path

import os
import numpy as np
import torch
import nixio as nix
import datetime
import sapicore

from tree_config.utils import get_class_bases
from tree_config.yaml import yaml_dumps
from tree_config import dump_config, read_config_from_file

__all__ = (
    "Loggable",
    "read_loggable_from_object",
    "get_loggable_properties",
    "read_loggable_from_file",
    "update_loggable_from_object",
    "dump_loggable",
    "log_tensor_board",
    "load_save_get_loggable_properties",
    "NixLogWriter",
    "NixLogReader",
)

FlatLoggableProps = List[Tuple[List[str], Any, str, bool]]
LogDataItem = Tuple[Any, str, nix.DataArray, nix.DataArray, nix.DataArray, nix.DataArray]


class Loggable:
    """The :class:`~utils.loggable.Loggable` can be used as a base class for objects that need to control which
    of its properties can be logged.

    When getting loggable properties, :func:`read_loggable_from_object` will call :meth:`loggable_props` and
    :meth:`loggable_children`, which looks up all the properties from :attr:`_loggable_props_` and
    :attr:`_loggable_children_`.

    Example
    -------
    The following class has two loggable properties:

    >>> class LoggableModel(Loggable):
    >>>     _loggable_props_ = ("frame", "color")
    >>>     frame: str = "square"
    >>>     color: str = "blue"

    This snippet will return what they are after instantiating an object of this class:

    >>> LoggableModel().loggable_props
    ['frame', 'color']

    To accumulate over loggable children as well:

    >>> read_loggable_from_object(LoggableModel(), True)
    {'frame': True, 'color': True}

    """

    _loggable_props_: Tuple[str] = ()
    _loggable_props: List[str] = None
    _loggable_children_: Dict[str, str] = {}
    _loggable_children: Dict[str, str] = None

    @property
    def loggable_props(self) -> List[str]:
        """A property containing a list of instance attribute names that are loggable (i.e., listed in
        `_loggable_props_` of this class and its parent classes)."""
        props = self._loggable_props
        if props is None:
            props = {}
            cls = self.__class__

            for c in [cls] + list(get_class_bases(cls)):
                if "_loggable_props_" not in c.__dict__:
                    continue

                for prop in c._loggable_props_:
                    if prop in props:
                        continue

                    if not hasattr(self, prop):
                        raise Exception("Missing attribute <{}> in <{}>".format(prop, cls.__name__))
                    props[prop] = None

            props = self._loggable_props = list(props)
        return props

    @property
    def loggable_children(self) -> Dict[str, str]:
        """A property containing a dictionary of loggable attribute names from derivative classes (i.e.,
            listed in `_loggable_children_` of this class and its parent classes).

        Example
        -------
        >>> class Thing(Loggable):
        >>>     _loggable_children_ = {"useful": "box"}
        >>>     box: str = None
        >>>
        >>> class Box(Loggable):
        >>>     pass
        >>>
        >>> model = Thing()
        >>> model.box = Box()
        >>> model.loggable_children
        {'useful': 'box'}

        """
        children = self._loggable_children
        if children is None:
            self._loggable_children = children = {}
            cls = self.__class__

            for c in [cls] + list(get_class_bases(cls)):
                if "_loggable_children_" not in c.__dict__:
                    continue

                for name, prop in c._loggable_children_.items():
                    if name in children:
                        continue

                    if not hasattr(self, prop):
                        raise Exception("Missing attribute <{}> in <{}>".format(prop, cls.__name__))
                    children[name] = prop

        return children


def _fill_loggable_from_declared_objects(
    classes: Dict[str, Any], loggable: Dict[str, Any], default_value: bool
) -> None:
    for name, obj in classes.items():
        if isinstance(obj, dict):
            loggable[name] = {k: read_loggable_from_object(o, default_value) for k, o in obj.items()}
        elif isinstance(obj, (list, tuple)):
            loggable[name] = [read_loggable_from_object(o, default_value) for o in obj]
        else:
            loggable[name] = read_loggable_from_object(obj, default_value)


def read_loggable_from_object(obj: Loggable, default_value: bool) -> Dict[str, Any]:
    """Returns a recursive dict containing all the loggable properties of the
    obj and its loggable children. Each item maps the item name to
    ``default_value``, indicating whether to log the item by default.

    :param obj: The object from which to get the loggables.
    :param default_value: A bool (True/False), indicating whether to default
        the loggable properties to True (they are logged) or False (they are
        not logged). The individual properties can subsequently be set either
        way manually in the dict.

    E.g.:

    .. code-block:: python

        class Model(Loggable):

            _loggable_props_ = ('name', )

            _loggable_children_ = {'the box': 'box'}

            name = 'chair'

            box = None

        class Box(Loggable):

            _loggable_props_ = ('volume', )

            volume = 12

    then:

    .. code-block:: python

        >>> model = Model()
        >>> model.box = Box()
        >>> loggables = read_loggable_from_object(model, False)
        >>> loggables
        {'the box': {'volume': False}, 'name': False'}
        >>> # to log name
        >>> loggables['name'] = True
    """
    # TODO: break infinite cycle if obj is listed in its nested loggable
    #  classes
    loggable = {}

    # get all the loggable classes used by the obj
    objects = {name: getattr(obj, prop) for name, prop in obj.loggable_children.items()}

    _fill_loggable_from_declared_objects(objects, loggable, default_value)

    for attr in obj.loggable_props:
        if attr not in loggable:
            loggable[attr] = default_value
    return loggable


def _get_loggable_from_declared_objects(
    property_path: List[str], classes: Dict[str, Any], friendly_names: Dict[str, str], loggable: Dict[str, Any]
) -> FlatLoggableProps:
    loggable_properties = []
    for name, obj in classes.items():
        if obj is None:
            continue

        friendly_name = friendly_names.get(name, name)
        if friendly_name in loggable:
            loggable_properties.extend(get_loggable_properties(obj, loggable[friendly_name], property_path + [name]))

    return loggable_properties


def get_loggable_properties(
    obj: Loggable, loggable: Dict[str, Any], property_path: Optional[List[str]] = None
) -> FlatLoggableProps:
    """Takes the loggable data read with :func:`read_loggable_from_object`
    or :func:`read_loggable_from_file`, filters those properties that should
    be logged (i.e. those that are True) and flattens them into a single flat
    list.

    This list could be used to efficiently log all the properties that need
    to be logged, by iterating the list.

    :param obj: The object from which to get the loggables.
    :param loggable: The loggable data previously read with
        :func:`read_loggable_from_object` or :func:`read_loggable_from_file`.
    :param property_path: A list of strings, indicating the object children
        names, starting from some root object that lead to the ``obj``.
        This can be used to reconstruct the full object path to each property
        starting from ``obj``.

        Should just be None (default) in user code.
    :returns: A list of 4-tuples, each containing
        ``(property_path, item, prop, value)``. ``property_path`` is the list
        of strings, indicating the object children names,
        starting from some root object that lead to the ``item``. ``item`` is
        the object or child whose property ``prop`` is being considered.
        ``value`` is a bool indicating whether to log the property.

    E.g.:

    .. code-block:: python

        class Model(Loggable):

            _loggable_props_ = ('name', )

            _loggable_children_ = {'the box': 'box'}

            name = 'chair'

            box = None

        class Box(Loggable):

            _loggable_props_ = ('volume', )

            volume = 12

    then:

    .. code-block:: python

        >>> model = Model()
        >>> model.box = Box()
        >>> d = read_loggable_from_object(model, False)
        {'the box': {'volume': False}, 'name': False'}
        >>> d['the box']['volume'] = True
        >>> get_loggable_properties(model, d)
        [(['the box', 'volume'], <Box at 0x16f0b2b5320>, 'volume', True)]
    """
    property_path = property_path or []
    loggable_properties = []
    # get all the loggable classes used by the obj
    objects = {prop: getattr(obj, prop) for name, prop in obj.loggable_children.items()}
    friendly_names = {v: k for k, v in obj.loggable_children.items()}
    props = set(obj.loggable_props)

    loggable_properties.extend(_get_loggable_from_declared_objects(property_path, objects, friendly_names, loggable))

    loggable_values = {k: v for k, v in loggable.items() if k not in objects and k in props}

    for k, v in loggable_values.items():
        if v:
            loggable_properties.append((property_path + [k], obj, k, v))

    return loggable_properties


def _update_loggable_from_declared_objects(
    classes: Dict[str, Any], loggable: Dict[str, Any], new_loggable: Dict[str, Any], default_value: bool
) -> None:
    for name, obj in classes.items():
        if isinstance(obj, dict):
            obj_loggable = loggable.get(name, {})
            new_loggable[name] = {
                k: update_loggable_from_object(o, obj_loggable.get(k, {}), default_value) for k, o in obj.items()
            }
        elif isinstance(obj, (list, tuple)):
            obj_loggable = loggable.get(name, [])
            new_loggable[name] = items = []
            for i, o in enumerate(obj):
                if i < len(obj_loggable):
                    items.append(update_loggable_from_object(o, obj_loggable[i], default_value))
                else:
                    items.append(update_loggable_from_object(o, {}, default_value))
        else:
            new_loggable[name] = update_loggable_from_object(obj, loggable.get(name, {}), default_value)


def update_loggable_from_object(obj: Loggable, loggable: Dict[str, Any], default_value: bool) -> Dict[str, Any]:
    """Takes the loggable data read with :func:`read_loggable_from_object`
    or :func:`read_loggable_from_file` and creates a new loggable dict from it
    and updates it by adding any new loggable properties of the object, that has
    been added to the object or its children since the last time it was
    created.

    :param obj: The object from which to get the loggables.
    :param loggable: The the loggable data read with
        :func:`read_loggable_from_object` or :func:`read_loggable_from_file`.
    :param default_value: A bool (True/False), indicating whether to default
        the new loggable properties to True (they are logged) or False (they are
        not logged). The individual properties can subsequently be set either
        way manually in the dict.
    :returns: The updated loggable dict.
    """
    # TODO: break infinite cycle if obj is listed in its nested loggable
    #  classes
    new_loggable = {}

    # get all the loggable classes used by the obj
    objects = {name: getattr(obj, prop) for name, prop in obj.loggable_children.items()}

    _update_loggable_from_declared_objects(objects, loggable, new_loggable, default_value)

    for attr in obj.loggable_props:
        if attr not in new_loggable:
            new_loggable[attr] = loggable.get(attr, default_value)
    return new_loggable


def read_loggable_from_file(filename: Union[str, Path]) -> Dict[str, Any]:
    """Reads and returns the dict indicating whether to log objects from a file
    that was previously dumped with :func:`dump_loggable`.

    :param filename: The filename to the yaml file.

    E.g.:

    .. code-block:: python

        >>> class Model(Loggable):
        >>>     _loggable_props_ = ('name', )
        >>>     name = 'chair'
        >>> model = Model()
        >>> d = read_loggable_from_object(model, True)
        >>> dump_loggable('loggable.yaml', d)
        >>> read_loggable_from_file('loggable.yaml')
        {'name': True}
    """
    return read_config_from_file(filename)


def dump_loggable(filename: Union[str, Path], data: Dict[str, Any]) -> None:
    """Dumps the loggable data gotten with e.g.
    :func:`read_loggable_from_object` to a yaml file.

    :param filename: The yaml filename.
    :param data: The loggable data.

    E.g.:

    .. code-block:: python

        >>> class Model(Loggable):
        >>>     _config_props_ = ('name', )
        >>>     name = 'chair'
        >>> model = Model()
        >>> d = read_loggable_from_object(model, True)
        >>> dump_config('loggable.yaml', d)
        >>> with open('loggable.yaml') as fh:
        ...     print(fh.read())

    Which prints ``name: true``.
    """
    dump_config(filename, data)


def log_tensor_board(
    writer,
    loggable_properties: FlatLoggableProps,
    global_step: Optional[int] = None,
    walltime: Optional[float] = None,
    prefix: str = "",
) -> None:
    """Logs the current value of all the given properties to the TensorBoard
    writer object for live display.

    Properties that are arrays are logged as a sequence of scalars, which is
    inefficient. It is suggested to use :class:`NixLogWriter` for complex
    data and :func:`log_tensor_board` only for scalars.

    :param writer: An instantiated ``torch.utils.tensorboard.SummaryWriter``
        object to which the data will be written.
    :param loggable_properties: The list of properties (in the format of
        :func:`get_loggable_properties`) to read and log.
    :param global_step: The optional global timestep, which must be
        monotonically increasing if providing.
    :param walltime: The optional wall time.
    :param prefix: An optional prefix used to log the data under.
    """
    for names, obj, prop, selection in loggable_properties:
        if not selection:
            continue

        value = getattr(obj, prop)
        if prefix:
            names = [prefix] + list(names)
        tag = "/".join(names)

        if isinstance(value, torch.Tensor):
            value = value.cpu().numpy()
        else:
            value = np.asarray(value)

        if value.shape:
            if isinstance(selection, (list, tuple)):
                flat_idx = np.ravel_multi_index(np.array(selection).T, value.shape)
                flat_val = np.take(np.ravel(value), flat_idx)
                items = {",".join(map(str, idx)): val for idx, val in zip(selection, flat_val)}
            else:
                items = {",".join(map(str, idx)): val for idx, val in np.ndenumerate(value)}
            writer.add_scalars(tag, items, global_step=global_step, walltime=walltime)
        else:
            writer.add_scalar(tag, value.item(), global_step=global_step, walltime=walltime)


def load_save_get_loggable_properties(obj: Any, filename: Union[str, Path], default_value: bool) -> FlatLoggableProps:
    """Loads the loggable configuration from the yaml file (
    :func:`read_loggable_from_file`, if the file doesn't
    exist it creates it with :func:`read_loggable_from_object` and
    :func:`dump_loggable`), updates it with :func:`update_loggable_from_object`
    and then dumps it back to the yaml file with :func:`dump_loggable`. It also
    returns the flattened loggable properties using
    :func:`get_loggable_properties`.

    This can be used to get the loggable properties, but also making sure the
    file contains the current loggables including any new properties not
    previously there.

    :param obj: The loggable object.
    :param filename: The yaml filename.
    :param default_value: A bool (True/False), indicating whether to default
        the new loggable properties to True (they are logged) or False (they are
        not logged). The individual properties can subsequently be set either
        way manually in the file.
    :returns: The loggable list returned by :func:`get_loggable_properties`.

    E.x.:

    .. code-block:: python

        class Model(Loggable):

            _config_props_ = ('name', )

            name = 'chair'

        class ModelV2(Loggable):

            _config_props_ = ('name', 'side')

            name = 'chair'

            side = 'left'

    then:

    .. code-block:: python

        >>> model = Model()
        >>> load_save_get_loggable_properties(model, 'loggable.yaml', True)
        [(['name'], <Model at 0x16f0b2b5278>, 'name', True)]
        >>> # then later for v2 of the model
        >>> model_v2 = ModelV2()
        >>> load_save_get_loggable_properties(model_v2, 'loggable.yaml', True)
        [(['name'], <ModelV2 at 0x16f0b2b5695>, 'name', True),
         (['side'], <ModelV2 at 0x16f0b2b5695>, 'side', True)]
        >>> with open('loggable.yaml') as fh:
        ...     print(fh.read())

    this prints::

        name: true
        side: true
    """
    if not os.path.exists(filename):
        dump_loggable(filename, read_loggable_from_object(obj, default_value))
    loggable = read_loggable_from_file(filename)
    dump_loggable(filename, update_loggable_from_object(obj, loggable, default_value))

    return get_loggable_properties(obj, loggable)


class NixLogWriter:
    """Writer that supports logging properties to a Nix HDF5 file.

    It supports logging arbitrary torch tensors, numpy arrays, and numerical
    scalars. The logged data can then be retrieved with :class:`NixLogReader`.

    A typical example is:

    . code-block:: python

        log_props = read_loggable_from_object(Model(), True)
        writer = NixLogWriter(h5_filename)
        # create the file
        writer.create_file()
        # create a section for this session
        writer.create_block('example')
        # create the objects required for logging using the loggable properties
        property_arrays = writer.get_property_arrays('example', log_props)
        do_work...
        # all the enabled loggable props will be logged
        writer.log(property_arrays, 0)
        do_more_work...
        writer.log(property_arrays, 1)
        # now close
        writer.close_file()
    """

    nix_file: Optional[nix.File] = None
    """The internal nix file object."""

    filename: Union[str, Path] = ""
    """The filename of the nix file that will be created when
    :meth:`create_file` is called.
    """

    git_hash = ""
    """An optional git hash that can be included in the file. It can be used
    to label the model version for example.
    """

    compression = nix.Compression.Auto
    """The compression to use in the file.

    It is one of ``nix.Compression``.
    """

    metadata: Optional[dict] = None
    """A dict of arbitrary metadata that can be included in the file.

    It is saved to the file after being encoded using
    :meth:`~tree_config.yaml.yaml_dumps`.
    """

    config_data: Optional[dict] = None
    """A dict of :mod:`tree_config` config data used to config the model that
    can be included in the file.

    It is saved to the file after being encoded using
    :meth:`~tree_config.yaml.yaml_dumps`.
    """

    def __init__(
        self,
        filename: Union[str, Path],
        git_hash="",
        compression=nix.Compression.Auto,
        metadata: Optional[dict] = None,
        config_data: Optional[dict] = None,
    ):
        self.filename = filename
        self.git_hash = git_hash
        self.compression = compression
        self.metadata = metadata
        self.config_data = config_data

    def __enter__(self):
        self.create_file()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_file()

    def create_file(self) -> None:
        """Creates the file named in :attr:`filename`. If it exists, an error
        is raised.

        :meth:`close_file` must be called after this to close the file.
        """
        if os.path.exists(self.filename):
            raise ValueError(f"{self.filename} already exists")

        self.nix_file = f = nix.File.open(str(self.filename), compression=self.compression, mode=nix.FileMode.Overwrite)

        sec = f.create_section("config", "configuration")
        sec["sapicore_version"] = sapicore.__version__
        sec["current_utc_time"] = str(datetime.datetime.utcnow())
        sec["git_hash"] = self.git_hash
        sec["config_data"] = yaml_dumps(self.config_data or {})

        metadata_sec = sec.create_section("metadata", "metadata")
        for key, value in (self.metadata or {}).items():
            metadata_sec[key] = yaml_dumps(value)

    def close_file(self) -> None:
        """Closes the nix file."""
        if self.nix_file is not None:
            self.nix_file.close()
            self.nix_file = None

    def create_block(self, name: str) -> None:
        """Creates a new block using the given that could subsequently be used
        with :meth:`get_property_arrays` to log to this block.

        It is used to organize the data into sessions.
        """
        self.nix_file.create_block(f"{name}_data_shape_len", "shape")
        self.nix_file.create_block(f"{name}_data_shape", "shape")
        self.nix_file.create_block(f"{name}_data_log", "data")
        self.nix_file.create_block(f"{name}_data_counter", "counter")

    def get_property_arrays(self, name: str, loggable_properties: FlatLoggableProps) -> List[LogDataItem]:
        """Takes the list of properties to be logged and it creates a list of
        objects that can be used with :attr:`log` to log all these properties.

        See the class for an example.

        :param name: The name of the block to which the properties will be
            logged. The block must have been created previously with
            :meth:`create_block`.
        :param loggable_properties: The list of properties (in the format of
            :func:`get_loggable_properties`) to read and log.
        :return: A list of objects that can be passed to :meth:`log` to log
            the properties.
        """
        shape_len_block: nix.Block = self.nix_file.blocks[f"{name}_data_shape_len"]
        shape_block = self.nix_file.blocks[f"{name}_data_shape"]
        data_block = self.nix_file.blocks[f"{name}_data_log"]
        counter_block = self.nix_file.blocks[f"{name}_data_counter"]

        property_data = []
        for names, obj, prop, selection in loggable_properties:
            if not selection:
                continue

            value = getattr(obj, prop)
            tag = ":".join(names)

            if isinstance(value, torch.Tensor):
                dtype = value.cpu().numpy().dtype
            else:
                if isinstance(value, (int, float)):
                    dtype = np.float64
                else:
                    raise ValueError(
                        f"Unrecognized data type for value <{value}>, with "
                        f"data type {type(value)} for {obj} property {prop}"
                    )

            if tag in data_block.data_arrays:
                data = data_block.data_arrays[tag]
                counter = counter_block.data_arrays[tag]
                shape_len = shape_len_block.data_arrays[tag]
                shape = shape_block.data_arrays[tag]
            else:
                data = data_block.create_data_array(tag, "data", dtype=dtype, data=[])
                counter = counter_block.create_data_array(tag, "counter", dtype=np.int64, data=[])
                shape_len = shape_len_block.create_data_array(tag, "shape_len", dtype=np.int8, data=[])
                shape = shape_block.create_data_array(tag, "shape", dtype=np.int32, data=[])

            property_data.append((obj, prop, counter, data, shape_len, shape))

        return property_data

    @staticmethod
    def log(property_arrays: List[LogDataItem], counter: int) -> None:
        """Logs all the properties passed to :meth:`get_property_arrays`.

        See the class for an example.

        :param property_arrays: The list of objects returned by
            :meth:`get_property_arrays`
        :param counter: An arbitrary integer that gets logged with the
            properties that could be later used to identify the data,
            e.g. the epoch when it was logged.
        """
        for obj, prop, counter_array, data, shape_len, shape in property_arrays:
            counter_array.append(counter)

            value = getattr(obj, prop)
            if isinstance(value, torch.Tensor):
                value = value.cpu().numpy()
            value = np.asarray(value)

            n = len(value.shape)
            shape_len.append(n)
            if n:
                shape.append(value.shape)
            data.append(np.ravel(value))


class NixLogReader:
    """A reader that can read and return the data written to a nix file using
    :class:`NixLogWriter`.

    E.g.:

    . code-block:: python

        with NixLogReader('log.h5') as reader:
            print('Logged experiments: ', reader.get_experiment_names())
            # prints e.g. `Logged experiments:  ['example']`
            print('Logged properties: ', reader.get_experiment_property_paths('example'))
            # prints e.g. `Logged properties:  [('neuron_1', 'activation'), ...`
            print('Activation: ', reader.get_experiment_property_data('example', ('neuron_1', 'activation')))
            # prints `Activation:  (array([0, 1, 2], dtype=int64), [array([0...`
    """

    nix_file: Optional[nix.File] = None
    """The internal nix file object."""

    filename: Union[str, Path] = ""
    """The filename of the nix file that will be opened when
    :meth:`open_file` is called.
    """

    def __init__(self, filename: Union[str, Path]):
        self.filename = filename

    def __enter__(self):
        self.open_file()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_file()

    def open_file(self):
        """Opens the file named in :attr:`filename` that was created with
        :meth:`NixLogWriter.create_file`.

        :meth:`close_file` must be called after this to close the file.

        An alternative syntax is to use it as a context manager, which is a
        little safer:

        . code-block:: python

            with NixLogReader('log.h5') as reader:
                ...

        This internally calls :meth:`open_file` and :meth:`close_file` as
        needed.
        """
        self.nix_file = nix.File.open(str(self.filename), mode=nix.FileMode.ReadOnly)

    def close_file(self):
        """Closes the nix file."""
        if self.nix_file is not None:
            self.nix_file.close()
            self.nix_file = None

    def get_experiment_names(self) -> List[str]:
        """Returns the list of block names found in the file that was created
        with :meth:`NixLogWriter.create_block`.

        E.g.:

        . code-block:: python

            >>> with NixLogReader('log.h5') as reader:
            >>>     print(reader.get_experiment_names())
            ['example']
        """
        f = self.nix_file
        return [block.name[:-9] for block in f.blocks if block.name.endswith("_data_log")]

    def get_experiment_property_paths(self, name: str) -> List[Tuple[str]]:
        """Returns the list of properties found in the file that was created
        with :meth:`NixLogWriter.get_property_arrays`.

        Each item in the list is itself a tuple of strings. This tuple
        represents the path to the property starting from the root object (e.g.
        the model) as attributes. E.g. ``model.child.prop`` will be represented
        as ``('model', 'child', 'prop')``.

        :param name: The name of the block to scan for properties.

        E.g.:

        . code-block:: python

            >>> with NixLogReader('log.h5') as reader:
            >>>     print(reader.get_experiment_property_paths('example'))
            [('neuron_1', 'activation'), ('neuron_1', 'intensity'),
            ('synapse', 'activation'), ('neuron_2', 'activation'),
            ('neuron_2', 'intensity'), ('activation_sum',)]
        """
        data_block: nix.Block = self.nix_file.blocks[f"{name}_data_log"]
        properties = []
        for arr in data_block.data_arrays:
            properties.append(tuple(arr.name.split(":")))

        return properties

    def get_experiment_property_data(
        self, name: str, property_path: Tuple[str, ...]
    ) -> Tuple[np.ndarray, List[Union[np.ndarray, float, int]]]:
        """Returns all the data of a property previously logged with
        :meth:`NixLogWriter.log`.

        :param name: The name of the block that contains the data.
        :param property_path: The full attribute path to the property, so we
            can locate it in the data. It is an item from the list returned by
            :meth:`get_experiment_property_paths`.
        :return: A 2-tuple of ``(counter, data)``. ``counter`` and ``data``,
            have the same length. ``counter`` is an array containing the count
            value that corresponds the each data item as passed to
            the ``counter`` parameter in :meth:`NixLogWriter.log`.
            ``data`` is a list of the values logged for this property, one item
            for each call to :meth:`NixLogWriter.log`. It can be scalars or
            numpy arrays.

        E.g.:

        . code-block:: python

            >>> with NixLogReader('log.h5') as reader:
            >>>     print(reader.get_experiment_property_data(
            ...           'example', ('neuron_1', 'activation')))
            (array([0, 1, 2], dtype=int64), [
                array([0.56843126, 1.0845224 , 1.3985955 ], dtype=float32),
                array([0.40334684, 0.83802634, 0.7192576 ], dtype=float32),
                array([0.40334353, 0.59663534, 0.18203649], dtype=float32)
            ])
        """
        shape_len_block: nix.Block = self.nix_file.blocks[f"{name}_data_shape_len"]
        shape_block = self.nix_file.blocks[f"{name}_data_shape"]
        data_block = self.nix_file.blocks[f"{name}_data_log"]
        counter_block = self.nix_file.blocks[f"{name}_data_counter"]

        tag = ":".join(property_path)

        counter = np.asarray(counter_block.data_arrays[tag])

        data = np.asarray(data_block.data_arrays[tag])
        shape_len = np.asarray(shape_len_block.data_arrays[tag])
        shape = np.asarray(shape_block.data_arrays[tag])

        items = [
            0,
        ] * len(shape_len)
        shape_s = 0
        data_s = 0
        for i, n in enumerate(shape_len):
            if not n:
                items[i] = data[data_s].item()
                data_s += 1
                continue

            dim = shape[shape_s : shape_s + n]
            k = np.multiply.reduce(dim)
            items[i] = np.reshape(data[data_s : data_s + k], dim)

            shape_s += n
            data_s += k

        return counter, items



"""Components handle universal parameters and logic common to neurons, synapses, and their derivatives."""
from typing import Any
import os

from torch import Tensor
from torch.nn import Module

from sapicore.utils.constants import DT
from sapicore.utils.loggable import Loggable
from sapicore.utils.sweep import Sweep
from sapicore.utils.io import save_yaml

from tree_config import Configurable, apply_config

__all__ = ("Component",)


class Component(Module, Configurable, Loggable):
    """Model component base class.

    Defines instance attributes and methods shared by all model components.
    Implements generic operations on loggable and configurable attributes.

    Parameters
    ----------
    identifier: str, optional
        Human-readable name for the component instance.

    configuration: dict, optional
        Configuration dictionary used during parameter initialization.

    device: str, optional
        Specifies a hardware device on which to process this object's tensors, e.g. "cuda:0".
        Must be a valid torch.device() string value. Defaults to "cpu".

    kwargs:
        Additional instance attributes that the user may set.

    See Also
    --------
    :class:`~tree_config.Configurable`
    :class:`~utils.loggable.Loggable`
    :class:`~utils.sweep.Sweep`

    """

    def __init__(self, identifier: str = None, configuration: dict = None, device: str = "cpu", **kwargs):
        super().__init__()

        # model-related common instance attributes.
        self.identifier = identifier
        self.configuration = configuration

        # hardware device on which to process this component.
        self.device = device

        # simulation-related tracking variables.
        self.simulation_step = 0
        self.dt = kwargs.pop("dt", DT)

    def configure(self, configuration: dict[str, Any] = None, log_destination: str = ""):
        """Applies a configuration to this object by adding the keys of `configuration` as instance attributes,
        initializing their values, and updating the `_config_props_` tuple to reflect the new keys.

        Also updates `_loggable_props_` for this object if the configuration includes the key "loggable".

        Parameters
        ----------
        configuration: dict or str, optional
            Parameters to be applied to this component, given as a dictionary.
            If None, this method can be used strictly for saving the configuration to YAML without changing it.

        log_destination: str, optional
            Path to a destination YAML file where the configuration will be saved.
            If not provided, the configuration is not saved to file.

        Note
        ----
        This method is generic and can be used without alterations by all derivative classes.

        The sub-dictionary `model` may include the special `sweep` key, which specifies settings for the
        heterogeneous initialization of certain parameters in the form of a dictionary.

        """
        # apply this object's configuration by adding or updating instance variables.
        if configuration:
            self.configuration = configuration
            apply_config(self, configuration)

            # override _loggable_properties_ if provided in the configuration dictionary.
            if configuration.get("loggable"):
                self._loggable_props = []
                self._loggable_props_ = configuration.get("loggable")

                # refresh the cache so that the new properties are added.
                _ = self.loggable_props

        # if `log_destination` was passed, save the configuration dictionary to that location.
        if log_destination and os.path.exists(log_destination):
            save_yaml(self.configuration, log_destination)

    def forward(self, data: Tensor) -> dict:
        """Processes an input, updates the state of this component, and advances the simulation by one step.

        Parameters
        ----------
        data: Tensor
            Input to be processed (e.g., added to a neuron's numeric state tensor `voltage`).

        Returns
        -------
        dict
            A dictionary whose keys are loggable attributes and whose values are their states as of this time step.
            For potential use by a :class:`~pipeline.simulation.SimpleSimulator` or any other
            :class:`~pipeline.Pipeline` script handling runtime operations.

        Raises
        ------
        NotImplementedError
            The forward method must be implemented by each derived class.

        """
        raise NotImplementedError

    def heterogenize(self, num_combinations: int, unravel: bool = True):
        """Edits configurable tensor attributes based on a sweep search dictionary if one was provided by the user
        within this object's `configuration` dictionary.

        Note
        ----
        If a `sweep` key is not present in `configuration`, this method will pass silently, retaining the existing
        configurable values. It can be invoked safely from generic methods (e.g.,
        :meth:`~engine.network.Network.build`).

        Parameters
        ----------
        num_combinations: int
            Total number of combinations to return.

        unravel: bool
            When the property is a 2D tensor, dictates whether combination values should be assigned
            to individual elements (True) or overwrite entire rows (False).

            The former behavior is appropriate for synapse attributes. It is achieved using `numpy.unravel_index`,
            which maps flattened-like indices to their nD equivalents given the shape of the array.

            The latter behavior is called for when values are themselves 1D vectors (e.g., oscillator frequencies).

        Warning
        -------
        This method mutates the underlying component object. To simply view the combinations, initialize a
        :class:`~utils.sweep.Sweep` object with your desired search space dictionary and number of combinations.

        """
        # safe to call if object has no sweep specification or empty configuration; method will return silently.
        if self.configuration and self.configuration.get("model", {}).get("sweep"):
            # the sweep object maintains a list of valid parameter combinations, each of which is a dictionary.
            sweep = Sweep(
                search_space=self.configuration.get("model", {}).get("sweep"), num_combinations=num_combinations
            )
            # modify this instance in-place.
            sweep.heterogenize(obj=self, unravel=unravel)

    def loggable_state(self) -> dict:
        """Returns a dictionary of this object's loggable properties and their states as of this simulation step.

        Returns
        -------
        dict
            Dictionary containing loggable property names and their values.

        """
        return {prop: getattr(self, prop) for prop in self.loggable_props}



"""Synapses connect neurons or ensembles to each other."""
from collections import deque
from typing import Callable

import torch
import numpy as np

from torch import tensor, Tensor
from torch.nn.init import xavier_uniform_

from sapicore.engine.component import Component
from sapicore.engine.neuron import Neuron

__all__ = ("Synapse",)


class Synapse(Component):
    """Static synapse base class.

    This class provides basic parameters and logic for connecting ensemble instances. Static synapses keep their
    initial weights throughout the simulation unless manually adjusted. Synapse instances have 2D weight and
    connectivity matrices, reflecting their role as layer-to-layer connectors.

    Synapses own the following parameters, over and above those inherited from :class:`~engine.component.Component`:

    Parameters
    ----------
    src_ensemble: Neuron
        Reference to the presynaptic (sending) ensemble.

    dst_ensemble: Neuron
        Reference to postsynaptic ensemble (receiving) ensemble.

    weights: Tensor
        2D weight matrix.

    connections: Tensor.
        2D connectivity mask matrix.

    weight_max: float or Tensor
        Positive weight value limit for any synapse element (applies throughout the simulation).

    weight_min: float or Tensor
        Negative weight value limit for any synapse element (applies throughout the simulation).

    delay_ms: float or Tensor
        Generic transmission delay, managed by the synapse object using queues.
        Transmission delays are useful for some temporal coding schemes.

    simple_delays: bool
        Whether transmission delay values from one presynaptic element are the same for all postsynaptic targets.
        Provides a speed-control tradeoff (one matrix multiplication vs. N vector multiplications).

    weight_init_method: Callable
        Weight initialization method from `torch.nn.init`.

    connections: Tensor
        2D binary integer mask matrix indicating which source unit -> destination unit connections are enabled.
        Loggable attribute.

    weights: Tensor
        2D float matrix storing a weight for each pair of source and destination ensemble units.
        Loggable attribute.

    output: Tensor
        1D tensor storing the result of W*x matrix multiplication in the current simulation step.
        Loggable attribute.

    """

    _config_props_: tuple[str] = ("weight_max", "weight_min", "delay_ms", "simple_delays")
    _loggable_props_: tuple[str] = ("weights", "connections", "output")

    # declare loggable instance attributes registered as torch buffers.
    connections: Tensor

    def __init__(
        self,
        src_ensemble: [Neuron] = None,
        dst_ensemble: [Neuron] = None,
        weights: [Tensor] = None,
        connections: [Tensor] = None,
        weight_max: float = 1000.0,
        weight_min: float = -1000.0,
        delay_ms: int = 0,
        simple_delays: bool = True,
        weight_init_method: Callable = xavier_uniform_,
        **kwargs,
    ):
        super().__init__(**kwargs)

        # model-related common instance attributes.
        self.learning = False

        self.src_ensemble = src_ensemble
        self.dst_ensemble = dst_ensemble

        try:
            self.matrix_shape = (dst_ensemble.num_units, src_ensemble.num_units)

        except AttributeError:
            self.matrix_shape = (1, 1)

        self.num_units = torch.prod(torch.tensor(self.matrix_shape), 0).item()

        self.weight_max = torch.zeros(self.matrix_shape, dtype=torch.float, device=self.device) + weight_max
        self.weight_min = torch.zeros(self.matrix_shape, dtype=torch.float, device=self.device) + weight_min
        self.weight_init_method = weight_init_method  # xavier_uniform by default.

        # [FAST] synaptic delay ring buffer initialization.
        self.delay_ms = delay_ms
        self.delay_buffer = None
        self.delay_indices = None

        # [LEGACY] whether transmission delays from one presynaptic element are shared among postsynaptic targets.
        self.simple_delays = simple_delays
        self.delay_queue = None
        self.delayed_data = None
        self.legacy = kwargs.pop("legacy", False)

        self._compute_delays()

        # container for data pulled from the queue in this simulation step (input as seen by `dst_ensemble`).
        self.delayed_data = torch.zeros(self.matrix_shape[1], dtype=torch.float, device=self.device)

        # float matrix containing synaptic weights.
        if weights is not None:
            if not isinstance(weights, Tensor):
                w_init = torch.ones(self.matrix_shape, device=self.device) * weights
            else:
                w_init = weights.to(self.device)
            self.register_buffer("weights", w_init)
        else:
            self.register_buffer(
                "weights",
                self.weight_init_method(tensor=torch.zeros(self.matrix_shape, dtype=torch.float, device=self.device)),
            )

        # binary connectivity matrix marking enabled/disabled connections (integer mask, all-to-all by default).
        # initialize connections from constructor input, if given.
        if connections is not None:
            self.register_buffer("connections", connections.to(self.device))
        else:
            self.register_buffer("connections", torch.ones_like(self.weights, dtype=torch.bool))

        # if given, apply connection strategy; all connections enabled by default.
        conn_mode = kwargs.pop("conn_mode", "all")
        if conn_mode:
            self.connect(conn_mode, kwargs.pop("conn_prop", 1))

        # output 1D tensor containing data for the destination ensemble.
        self.register_buffer("output", torch.zeros(self.matrix_shape[0], dtype=torch.float, device=self.device))

        # clamp weights to acceptable range given this synapse's min and max.
        self.weights = torch.clamp(self.weights, self.weight_min, self.weight_max)

        # if synapse represents a connection from an ensemble to itself, zero out diagonal of mask matrix.
        if self.src_ensemble is self.dst_ensemble:
            self.connections.fill_diagonal_(0)

        # if autograd parameter provided to constructor at build time, turn autograd on or off for the weights.
        if hasattr(self, "autograd"):
            self.weights.requires_grad_(self.autograd)

        # specifies which loggable attribute should be considered this unit's output.
        self.output_field = "output"

    def _compute_delays(self):
        """Compute synaptic delay buffers. Backward compatible with Sapicore < 0.4.0.

        Note
        ----
        Invoked in :meth:`heterogenize`, as it can potentially alter delay values after synapse instantiation.

        """
        # initialize default transmission delay values as a tensor of size `src_ensemble.num_units`.
        self.delay_ms = torch.zeros(self.matrix_shape[1], dtype=torch.float, device=self.device) + self.delay_ms

        if self.legacy:
            if self.simple_delays:
                # maintains delayed data queues.
                self.delay_queue = [
                    deque(torch.zeros(delay, device=self.device)) for delay in (self.delay_ms // self.dt).int()
                ]
            else:
                # the extensive delay setting, where delayed input views may vary across postsynaptic elements.
                self.delay_ms = torch.zeros(self.matrix_shape, dtype=torch.float, device=self.device) + self.delay_ms
                self.delayed_data = torch.zeros(self.matrix_shape, dtype=torch.float, device=self.device)
                self.delay_queue = [
                    deque(torch.zeros(delay.int(), device=self.device))
                    for delay in (self.delay_ms / self.dt).flatten().int()
                ]
        else:
            delay_ms_int = self.delay_ms.int()
            self.max_delay = delay_ms_int.max().item() + 1

            self.delay_indices = self.delay_ms.clone().to(self.device).long()
            self.delay_buffer = torch.zeros(
                (self.matrix_shape[1], self.max_delay), dtype=torch.float, device=self.device
            )

    def connect(self, mode: str = "all", prop: float = 1.0):
        """Applies a predefined connectivity mask strategy.

        Parameters
        ----------
        mode: str
            Sapicore provides four built-in connectivity mask initialization options: "all", "one", "prop", "rand".
            "all" enables all connections (matrix of 1s) except self-connections (diagonal) if the destination and
            source ensemble are the same object. "one" connects the i-th source neuron to the i-th destination neuron,
            zeroing out all matrix elements **except** the diagonal. "prop" connects each source neuron to
            `prop`*`dst_ensemble.num_units` randomly selected destination neurons. "rand" sets a proportion `prop`
            of the matrix elements to 1 with no balance constraints (unlike "prop").

        prop:
            Desired proportion of neurons in destination receiving a connection from a particular neuron
            in the source ensemble. Rounded to an integer using np.round.

            E.g., if prop = 0.2 and matrix_shape is (100, 5), 20 random row elements (destination neurons)
            in the connectivity mask will be set to 1.

        Raises
        ------
        ValueError
            If `mode` was set to a value other than "all", "one", "prop", or "rand".

        Warning
        -------
        Attempting to initialize one-to-one connectivity between ensembles of different sizes will necessarily
        result in some unconnected neurons. The faux-diagonal will be set to 1s but there will be all 0s rows/cols,
        depending on whether the source or destination ensemble is bigger.

        """
        match mode:
            case "all":
                self.connections = torch.ones_like(self.connections, device=self.device, dtype=torch.bool)

            case "one":
                num_src = self.src_ensemble.num_units
                num_dst = self.dst_ensemble.num_units

                # produces identity-like non-square matrices if applicable.
                self.connections = torch.eye(num_src, device=self.device, dtype=torch.bool).repeat_interleave(
                    num_dst // num_src, dim=0
                )

            case "prop":
                selection_size = int(np.round(prop * self.matrix_shape[0]))
                torch.randperm(self.matrix_shape[0])

                self.connections = torch.zeros_like(self.connections, device=self.device, dtype=torch.bool)
                for i in range(self.matrix_shape[1]):
                    self.connections[torch.randperm(self.matrix_shape[0])[:selection_size], i] = 1

            case "rand":
                num_enabled = int(np.round(prop * self.weights.numel()))
                ids_enabled = np.random.choice(np.arange(self.weights.numel()), size=num_enabled, replace=False)

                self.connections = torch.zeros_like(self.connections, device=self.device, dtype=torch.bool)
                self.connections[np.unravel_index(ids_enabled, shape=self.matrix_shape)] = 1.0

            case _:
                raise ValueError(f"Incorrect mode specified {mode}. Use 'all', 'one', 'prop', or 'rand'.")

    def heterogenize(self, num_combinations: int, unravel: bool = True):
        """Diversifies parameters in accordance with this synapse configuration dictionary and recomputes
        the spike delay queue in case values were altered."""
        super().heterogenize(num_combinations=self.num_units, unravel=unravel)

        # recompute delay steps and reinitialize ring buffer (or queues in legacy).
        self._compute_delays()

    def set_learning(self, state: bool = False) -> None:
        """Switch weight updates on or off, e.g. before a training/testing round commences.

        Parameters
        ----------
        state: bool
            Toggle learning on if True, off if False.

        Note
        ----
        Fine-grained control over which synapse elements are toggled on will be added in the future, to support
        more sophisticated algorithms. Currently, the global learning switch is meant to be used, e.g.,
        when feeding test buffer to a trained network with STDP synapses.

        """
        self.learning = state

    def set_weights(self, weight_initializer: Callable, *args, **kwargs) -> None:
        """Initializes 2D weight matrix for this synapse instance.

        Parameters
        ----------
        weight_initializer: Callable
            Any weight initialization method from :mod:`torch.nn.init`. Args and kwargs are passed along to it.

        """
        self.weights = weight_initializer(tensor=self.weights, *args, **kwargs)

    def queue_input(self, current_data: Tensor) -> Tensor:
        """Loads incoming spikes into the back of a ring buffer at the correct index,
        and returns the delayed spikes ready for transmission (from the front of the queue).

        Note
        ----
        Legacy variant uses a list of queues to implement transmission delays and "release" input appropriately.

        Initializes a list whose Nth element is a queue pre-filled with ``self.delay_ms[N] // DT`` zeros.
        On each forward iteration, call this method. It will enqueue the current source input data value and
        :meth:`popleft` the head of the queue.

        Returns
        -------
        Tensor
            1D tensor of input that occurred an appropriate number of steps ago for each unit.

        """
        if not self.legacy:
            # Add the incoming spikes at the current positions indicated by delay_indices
            n_synapses = self.matrix_shape[1]
            self.delay_buffer[torch.arange(n_synapses), self.delay_indices % self.max_delay] = current_data

            # Transmit the spikes at index 0 (front of the queue)
            spikes_to_transmit = self.delay_buffer[:, 0]

            # Rotate the delay buffer to the left (advance the spikes)
            self.delay_buffer = torch.cat(
                [self.delay_buffer[:, 1:], torch.zeros((n_synapses, 1), device=self.device)], dim=1
            )
            return spikes_to_transmit

        else:
            if self.simple_delays:
                # append updated presynaptic data to the queues, each corresponding to one presynaptic element.
                for i, value in enumerate(current_data):
                    if i >= len(self.delay_queue):
                        breakpoint()
                    self.delay_queue[i].append(value)

            else:
                # append updated presynaptic data to the delayed view of each postsynaptic element (src X dst queues).
                for i in range(len(self.delay_queue)):
                    multi_index = np.unravel_index(i, self.matrix_shape)
                    self.delay_queue[i].append(current_data[multi_index[1]])

            # return appropriate data tensor while popping the head of the line.
            valid_data = [item.popleft() for item in self.delay_queue]
            return tensor(valid_data, device=self.device)

    # child classes would typically only implement one or more of the methods below this line.
    def update_weights(self) -> Tensor:
        """Static weight update rule. Placeholder for plastic synapse derivative classes (e.g., STDP)."""
        pass

    def forward(self, data: Tensor) -> dict:
        """Static synapse forward method. Amounts to a simple matrix multiplication.

        When called with synapses that have nonzero transmission delays, variably-old data will be used as
        dictated by the queue maintained by this instance.

        Parameters
        ----------
        data: Tensor
            Output vector from the input layer in the previous iteration (`spiked` buffer, containing 0s and 1s).

        Returns
        -------
        dict
            Dictionary containing `weights`, `connections`, and `output` for further processing.

        Note
        ----
        The simple transmission delay implementation, for runtime efficiency, assumes that recipient synapse elements
        all have a shared "view" of incoming activity in the presynaptic ensemble. This means a single matrix
        multiplication (dst x src) by (src x 1) = (dst x 1) is performed. A smaller queue list of length
        `src_ensemble.num_units` is maintained, as opposed to `src_ensemble.matrix_shape`.

        If `simple_delays` is toggled **off**, each neuron in the recipient ensemble may have a different delayed
        view of the source ensemble activity. This means that N vector multiplications must be performed:
        sig(W[k,i]*dd[i]) where N is the number of elements in the source ensemble, k the index of a destination
        element, i the index of a source element, W a weight, and dd the delayed data transmitted from the
        presynaptic neuron i to the postsynaptic neuron k.

        """
        # use a list of queues to track transmission delays and "release" input after the fact.
        self.delayed_data = self.queue_input(data)

        # enforce weight limits.
        self.weights = torch.clamp(self.weights, self.weight_min, self.weight_max)

        if self.simple_delays:
            # mask non-connections (repeated on every iteration in case mask was updated during simulation).
            self.output = torch.matmul(self.weights * self.connections, self.delayed_data)

        else:
            self.delayed_data = self.delayed_data.reshape(self.matrix_shape)

            # this loop is unavoidable, as N vector multiplications with different delayed_data are needed.
            for i in range(self.matrix_shape[0]):
                self.output[i] = torch.matmul(self.weights[i, :] * self.connections[i, :], self.delayed_data[i, :])

        # advance simulation step and return output dictionary.
        self.simulation_step += 1

        return self.loggable_state()



"""Configurable workflows and scripts.

Pipelines are flexible objects consisting of two logical parts: configurable parameters and a custom
script defined by overriding the abstract method :meth:`~Pipeline.run`.

To facilitate the development and testing of models, Sapicore provides a compact default simulation
pipeline with a sample configuration YAML. Advanced users may use those as a basis for custom workflows.

"""
from tree_config import Configurable
from sapicore.utils.io import load_apply_config

__all__ = ("Pipeline",)


class Pipeline(Configurable):
    """Pipeline base class.

    Parameters
    ----------
    config_or_path: dict or str, optional
        Dictionary or path to YAML to use with :meth:`apply_config`.

    Raises
    ------
    FileNotFoundError
        If provided with an invalid string `configuration` path.

    """

    def __init__(self, config_or_path: dict | str = None, **kwargs):
        if isinstance(config_or_path, str):
            config_or_path = load_apply_config(config_or_path, apply_to=self)

        if isinstance(config_or_path, dict):
            # reached whether it was transformed to a dictionary in the conditional above or provided as such.
            self.configuration = config_or_path

    def run(self) -> None:
        """Pipeline implementation, to be overridden by the user."""
        raise NotImplementedError



""" Spiking neurons emit all-or-none events once their numeric state surpasses a critical threshold.

Spiking neurons inherit all properties of their parent class :class:`~engine.neuron.Neuron`. In addition to `voltage`,
they also own the binary integer tensor `spiked`, which maintains their emitted action potentials.

"""
import torch
from torch import Tensor

from sapicore.engine.neuron import Neuron

__all__ = ("SpikingNeuron",)


class SpikingNeuron(Neuron):
    """Generic spiking neuron.

    Defines instance attributes shared by all derived spiking neuron classes.

    Parameters
    ----------
    spiked: Tensor
        A binary representation of spike events, registered as a PyTorch buffer.

    """

    _loggable_props_: tuple[str] = ("input", "voltage", "spiked")

    # declare instance attributes to be registered as pytorch buffers.
    spiked: Tensor

    def __init__(self, **kwargs):
        # register universal loggables and configurables using the parent method(s).
        super().__init__(**kwargs)

        # state-related attributes unique to spiking neurons, over and above base components and neurons.
        self.register_buffer("spiked", torch.zeros(1, dtype=torch.int8, device=self.device))

        # update preferred output field to spiked.
        self.output_field = "spiked"

    def forward(self, data: Tensor) -> dict:
        """Processes an input, updates the state of this component, and advances the simulation by one step.

        Parameters
        ----------
        data: Tensor
            Input to be processed (e.g., added to a neuron's numeric state tensor `voltage`).

        Returns
        -------
        dict
            A dictionary whose keys are loggable attributes and whose values are their states as of this time step.
            For potential use by a :class:`~pipeline.simulation.SimpleSimulator` or any other
            :class:`~pipeline.Pipeline` script handling runtime operations.

        Raises
        ------
        NotImplementedError
            The forward method must be implemented by each derived class.

        """
        raise NotImplementedError



""" Networks are graph representations of neuron ensembles connected by synapses. """
from typing import Sequence

import os
from itertools import compress

import networkx as nx
from networkx import DiGraph

import torch
from torch import Tensor
from torch.nn import Module

from sapicore.engine.component import Component
from sapicore.engine.ensemble import Ensemble
from sapicore.engine.neuron import Neuron
from sapicore.engine.synapse import Synapse

from sapicore.utils.constants import SYNAPSE_SPLITTERS
from sapicore.utils.io import DataAccumulatorHook, flatten, load_yaml, MonitorHook

import matplotlib.pyplot as plt

__all__ = ("Network",)


class Network(Module):
    """Generalized spiking neural network instance. Maintains and orchestrates the processing of multiple
    ensembles connected by synapses.

    Networks may be constructed from a configuration dictionary (read from a YAML or defined programmatically)
    by applying :meth:`build`. User may also initialize an empty network and apply :meth:`add_ensembles`
    and :meth:`add_synapses` to build the network `graph` directly. One or more `root` nodes may be specified
    in the YAML/dictionary or passed to the Network constructor.

    Parameters
    ----------
    identifier: str, optional
        String identifier for file I/O and visualization operations.

    configuration: dict, optional
        Configuration dictionary containing model specification and simulation parameters. Used to construct
        the `graph` and initialize the `root` instance attributes if provided.

    roots: list of str, optional
        List of one or more ensemble identifiers corresponding to root nodes, i.e. the starting point of each
        forward pass. Can be passed to the constructor, configured in a YAML (model/roots), or automatically
        detected by checking for nodes with in-degree=0 (the latter is a fallback strategy and does not override
        user configuration). In cases where multiple external data tensors are passed to
        :meth:`~engine.network.Network.forward` or :meth:`~model.Model.fit`, they are routed to these ensembles
        (the i-th tensor to the i-th root by convention).

    device: str
        Specifies a hardware device on which to process this object's tensors, e.g. "cuda:0".
        Must be a valid torch.device() string value. Defaults to "cpu".

    Note
    ----
    Synapse objects know their source and destination ensemble references; they are 2D objects
    capable of connecting two ensembles in a unidirectional fashion with heterogeneous parameters.
    Lateral connections from an ensemble to itself are supported by design and warrant no special treatment.

    On :meth:`forward` calls, the nodes of the graph are traversed in breadth-first search (BFS) order. The ensemble
    referenced by the current node is processed first, followed by its outgoing synapses. If multiple source nodes
    exist, the ``networkX`` multi-source BFS implementation is used.

    """

    def __init__(
        self, identifier: str = None, device: str = "cpu", configuration: dict = None, roots: list[str] = None
    ):
        super().__init__()

        # model-related common instance attributes.
        self.identifier = identifier
        self.configuration = configuration
        self.device = device
        self.graph = DiGraph()

        # if provided, ensure user's root node selection is preserved.
        self.roots = [] if roots is None else roots
        self.root_lock = True if self.roots else False

        # simulation-related common instance attributes.
        self.simulation_step = 0
        self.traversal_order = []

        # store identifiers of registered components for lookup purposes.
        self._ensemble_identifiers = []
        self._synapse_identifiers = []

        # network construction from configuration file overrides programmatic initialization.
        if configuration:
            self._build()

        # if still not set, automatically mark root nodes by their in-degree.
        if not self.root_lock:
            # keep unlocked in case of subsequent `add_ensemble` or `add_synapse` calls.
            self._find_roots()

    def __getitem__(self, item: str) -> Ensemble | Synapse | None:
        """Look up and return a network component by its string identifier."""
        if self.graph.nodes.get(item):
            return self.graph.nodes.get(item)["reference"]

        else:
            syn_splitter = list(compress(SYNAPSE_SPLITTERS, [syn_char in item for syn_char in SYNAPSE_SPLITTERS]))
            if len(syn_splitter) == 1:
                # the component name is guaranteed to contain one valid synapse splitter ("--" or "->").
                syn_splitter = syn_splitter[0]

            elif len(syn_splitter) > 1:
                # the component name is ambiguous because it contains more than one valid synapse splitter string.
                raise KeyError(
                    f"Invalid synapse identifier format {item}. Must only contain one of the following"
                    f"patterns: {SYNAPSE_SPLITTERS}"
                )

            else:
                # if the component referenced by 'item' is not a node or a valid synapse.
                raise KeyError(f"Component {item} does not exist in this network.")

            return self.graph.edges.get(item.split(syn_splitter))["reference"]

    def __str__(self):
        """Describe the network to the user."""
        num_neurons = sum([self.graph.nodes[i]["reference"].num_units for i in self.graph.nodes])

        total_synapses = int(sum([s[0] * s[1] for s in [i.matrix_shape for i in self.get_synapses()]]))
        active_synapses = int(sum([conn.sum() for conn in [i.connections for i in self.get_synapses()]]))

        return (
            f"'{self.identifier}': {len(self.graph.nodes)} layers, {num_neurons} neurons, "
            f"{len(self.graph.edges)} synapse matrices with {active_synapses}/{total_synapses} active connections."
        )

    def add_ensembles(self, *args: Neuron | dict | str):
        """Adds ensemble nodes to the network graph from paths to YAML, from pre-initialized
        :class:`engine.neuron.Neuron` objects or children thereof (e.g., any `Ensemble`), or from
        configuration dictionaries. User may mix and match if necessary.

        Raises
        ------
        TypeError
            If user attempts to pass an argument that is not a string path, a dictionary, or an instance derived
            from :class:`neuron.Neuron`.

        Warning
        -------
        The component `identifier` field is used to reference nodes and edge sources or destinations when
        constructing the network graph.

        """
        for ensemble in args:
            if isinstance(ensemble, str):
                if os.path.exists(ensemble):
                    # instantiate the ensemble from configuration file.
                    ensemble, _ = self._object_from_configuration(path=ensemble, comp_type="ensemble")
                else:
                    raise IOError(f"Could not add ensemble, invalid path given: {ensemble}")

            elif isinstance(ensemble, dict):
                ensemble, _ = self._object_from_configuration(cfg=ensemble, comp_type="ensemble")

            # verify that `ensemble` now references a valid object derived from Neuron.
            if not isinstance(ensemble, Neuron):
                raise TypeError("Ensembles must be given as object references, dictionaries, or paths to YAML.")

            # add ensemble to the network graph and register its identifier.
            if self._unique_name(ensemble.identifier):
                self.graph.add_node(ensemble.identifier, reference=ensemble)
                self._ensemble_identifiers.append(ensemble.identifier)
            else:
                raise ValueError(f"Ensemble named {ensemble.identifier} already exists in this network.")

            # heterogenize modifies the object iff a `sweep` specification was provided in the model configuration.
            ensemble.heterogenize(num_combinations=ensemble.num_units)

        # update root node list.
        if not self.root_lock:
            self._find_roots()

    def add_synapses(self, *args: Synapse | dict | str):
        """Adds synapse edges to the network graph from paths to YAML, from pre-initialized
        :class:`engine.synapse.Synapse` objects or children thereof (e.g.,`STDPSynapse`), or from
        configuration dictionaries. User may mix and match if necessary.

        Raises
        ------
        TypeError
            If user attempts to pass an argument that is not a string path, a dictionary, or an instance derived
            from :class:`synapse.Synapse`.

        """
        for synapse in args:
            if isinstance(synapse, str):
                # instantiate synapse from configuration file.
                if os.path.exists(synapse):
                    # add the synapse reference to the graph as a node.
                    synapse, _ = self._object_from_configuration(path=synapse, comp_type="synapse")
                else:
                    raise IOError(f"Could not add synapse, invalid path given: {synapse}")

            elif isinstance(synapse, dict):
                # instantiate synapse from dictionary.
                synapse, _ = self._object_from_configuration(cfg=synapse, comp_type="synapse")

            # verify that `synapse` now references a valid Synapse object and has source and destination ensembles.
            if isinstance(synapse, Synapse):
                if not (synapse.src_ensemble and synapse.dst_ensemble):
                    raise AttributeError(
                        f"Synapse {synapse.identifier} is missing a source and/or a destination "
                        f"ensemble attribute(s)."
                    )

                # add synapse to network graph and register its identifier.
                if self._unique_name(synapse.identifier):
                    self.graph.add_edge(
                        synapse.src_ensemble.identifier, synapse.dst_ensemble.identifier, reference=synapse
                    )
                    self._synapse_identifiers.append(synapse.identifier)
                else:
                    raise ValueError(f"Synapse named {synapse.identifier} already exists in this network.")

            else:
                raise TypeError("Synapses must be given as object references, dictionaries, or paths to YAML.")

            # heterogenize modifies the object iff a `sweep` specification was provided in the model configuration.
            synapse.heterogenize(num_combinations=synapse.num_units)

        # update root node list.
        if not self.root_lock:
            self._find_roots()

    def get_ensembles(self):
        """Returns all ensemble references in the network `graph`."""
        return [self.graph.nodes[i]["reference"] for i in self.graph.nodes]

    def get_synapses(self):
        """Returns all synapse references in the network `graph`."""
        return [self.graph.edges[s, d]["reference"] for s, d in self.graph.edges]

    def _build(self):
        """Adds and initializes named components to the graph from this instance's `configuration` dictionary.

        Note
        ----
        Network YAMLs reference ensemble and synapse YAMLs by their path relative to that specified in the `root`
        field (or the project root if that field was not provided).

        Warning
        -------
        This function is only called once during instantiation from a configuration file.
        Do not call it from external modules.

        """
        # set network identifier string from configuration file.
        self.identifier = self.configuration.get("identifier", "")

        # avoid repeatedly recomputing the roots while the network is being automatically built.
        self.root_lock = True

        # use network API to add ensembles and synapses based on configuration dictionary or YAML paths.
        if "ensembles" in self.configuration.get("model", {}):
            ensemble_paths = [
                os.path.join(self.configuration.get("root", ""), ep)
                for ep in self.configuration.get("model", {}).get("ensembles", ())
            ]
            self.add_ensembles(*ensemble_paths)

        if "synapses" in self.configuration.get("model", {}):
            synapse_paths = [
                os.path.join(self.configuration.get("root", ""), sp)
                for sp in self.configuration.get("model", {}).get("synapses", ())
            ]
            self.add_synapses(*synapse_paths)

        # set roots from configuration if key `roots` exists and roots were not set yet.
        config_roots = self.configuration.get("model", {}).get("roots", False)

        if config_roots:
            # update root nodes from configuration and lock s.t. it does not get recomputed.
            self.roots = config_roots

        # this will only be arrived at if user hasn't passed or configured `roots`.
        # note that self.roots is initialized to [] to prevent BFS failing on empty root list.
        elif not self.roots:
            # automatically mark nodes with in-degree == 0 as roots.
            self._find_roots()

            # unlock to allow automatic root recomputing in subsequent `add_ensembles` or `add_synapses` calls.
            self.root_lock = False

    def add_monitor_hook(
        self,
        steps: int = None,
        attrs: Sequence[str] = None,
        comps: Sequence[Component] = None,
        dtype: torch.dtype = torch.float,
    ) -> dict:
        """Attach a forward hook to some or all network components, buffering accumulated output in memory.

        Parameters
        ----------
        steps: int
            Total number of simulation steps in the experiment to be logged.
            Required for HDF5 sizing and chunk management.

        attrs: Sequence of str
            Attributes to log for the given components.

        comps: Sequence of Component, optional
            Components to attach data hooks to. If not provided, data will be logged for all components.

        dtype: torch.dtype
            Torch datatype for this monitor hook; initialize correctly to save memory.

        """
        hooks = {}
        if not comps:
            for ensemble in self.get_ensembles():
                hooks[ensemble.identifier] = MonitorHook(
                    ensemble, ensemble.loggable_props if not attrs else attrs, steps
                )
            for synapse in self.get_synapses():
                hooks[synapse.identifier] = MonitorHook(
                    synapse, synapse.loggable_props if not attrs else attrs, steps, dtype
                )
        else:
            for comp in comps:
                hooks[comp.identifier] = MonitorHook(comp, comp.loggable_props if not attrs else attrs, steps, dtype)

        return hooks

    def add_data_hook(
        self, data_dir: str, steps: int, attrs: Sequence[str] = None, comps: Sequence[Component] = None
    ) -> list:
        """Attach a data accumulator forward hook to some or all network components, logging data to disk.

        Parameters
        ----------
        data_dir: str
            Path to directory in which to save intermediate simulation output.

        steps: int
            Total number of simulation steps in the experiment to be logged.
            Required for HDF5 sizing and chunk management.

        attrs: Sequence of str
            Attributes to log for the given components.

        comps: Component, optional
            Components to attach data hooks to. If not provided, data will be logged for all components.

        """
        hooks = []
        if not comps:
            for ensemble in self.get_ensembles():
                hooks.append(
                    DataAccumulatorHook(ensemble, data_dir, ensemble.loggable_props if not attrs else attrs, steps)
                )

            for synapse in self.get_synapses():
                hooks.append(
                    DataAccumulatorHook(synapse, data_dir, synapse.loggable_props if not attrs else attrs, steps)
                )
        else:
            for comp in comps:
                hooks.append(DataAccumulatorHook(comp, data_dir, comp.loggable_props if not attrs else attrs, steps))

        return hooks

    def backward(self) -> None:
        """Processes a backward sweep for this network object.

        This is not required for SNNs that learn with STDP and can be implemented by child classes that
        might require a backward pass.

        """
        pass

    def forward(self, data: Tensor | list[Tensor]) -> dict:
        """Processes current simulation step for this network object.

        In this generic implementation, forward call order is determined by a BFS traversal starting from
        the edges fanning out of the root nodes. Every node is visited once; for each node, every outgoing synapse
        is forwarded. That way, we are guaranteed to forward any component exactly once on every simulation step.

        Parameters
        ----------
        data: Tensor or list of Tensor
            External input to be processed by this generic network.

        Returns
        -------
        dict
            Dictionary containing the loggable properties of this network's ensembles and synapses in this
            simulation step.

        Warning
        -------
        Networks now support multiple root nodes and external input streams. In those cases, the graph is
        traversed layer-wise, using multi-source BFS.

        See Also
        --------
        :meth:`simulation.SimpleSimulator.run`
            For a demonstration of how to programmatically feed heterogeneous external current to multiple nodes.
            Note that this functionality will eventually be relegated to the data loader, with arbitrary current
            injections during the simulation being implemented by biases.

        """
        # follow outgoing synapses between vertices, forwarding the entire network on each iteration.
        for i, ensemble in enumerate(self.traversal_order):
            # list all synapses coming in and out of this ensemble. Networkx references it by its string identifier.
            incoming_synapses = self._in_edges(ensemble)
            outgoing_synapses = self._out_edges(ensemble)

            # shortcut to actual ensemble reference.
            ensemble_ref = self.graph.nodes[ensemble]["reference"]

            if ensemble_ref.identifier not in self.roots:
                # apply an aggregation function to synaptic data flowing into this ensemble.
                if incoming_synapses:
                    inputs = [synapse.output for synapse in incoming_synapses]
                    ids = [synapse.identifier for synapse in incoming_synapses]

                    # aggregation is (micro)managed at the neuron level; torch.sum is used by default.
                    integrated_data = ensemble_ref.aggregate(inputs, identifiers=ids).to(self.device)

                else:
                    integrated_data = ensemble_ref.input

            else:
                # if this is a root node, treat its stream from the data loader as inbound synaptic input.
                external = [data[self.roots.index(ensemble_ref.identifier)]] if isinstance(data, list) else [data]
                feedback = [synapse.output for synapse in incoming_synapses]

                ids = [f"ext{z}" for z in range(len(external))] + [synapse.identifier for synapse in incoming_synapses]
                integrated_data = ensemble_ref.aggregate(external + feedback, identifiers=ids)

            # forward current ensemble.
            ensemble_ref(integrated_data)

            # use the ensemble's declared output field (voltage or spiked) to forward its outgoing synapses.
            [synapse(getattr(ensemble_ref, ensemble_ref.output_field).float()) for synapse in outgoing_synapses]

        self.simulation_step += 1

        # build and pass a state dictionary for use by calling module.
        state = {"ensembles": {}, "synapses": {}}
        for comp in self.get_ensembles():
            state["ensembles"][comp.identifier] = comp.loggable_state()

        for comp in self.get_synapses():
            state["synapses"][comp.identifier] = comp.loggable_state()

        return state

    def draw(self, path: str, node_size: int = 750):
        """Saves an SVG networkx graph plot showing ensembles and their general connectivity patterns.

        Parameters
        ----------
        path: str
            Destination path for network figure.

        node_size: int, optional
            Node size in network graph plot.

        Note
        ----
        May be extended and/or moved to a dedicated visualization package in future versions.

        """
        plt.figure()
        nx.draw(self.graph, node_size=node_size, with_labels=True, pos=nx.kamada_kawai_layout(self.graph))

        plt.savefig(fname=os.path.join(path, self.identifier + ".svg"))
        plt.clf()

    def _in_edges(self, node: str) -> list[Synapse]:
        """Returns list of synapse objects going into the ensemble `node`."""
        if node in list(self.graph.nodes):
            return [ref["reference"] for _, _, ref in list(self.graph.in_edges(node, data=True))]

    def _out_edges(self, node: str) -> list[Synapse]:
        """Returns list of synapse objects fanning out of the ensemble `node`."""
        if node in list(self.graph.nodes):
            return [ref["reference"] for _, _, ref in list(self.graph.out_edges(node, data=True))]

    def _unique_name(self, candidate: str) -> bool:
        """Checks whether a component identifier is identical to one that has already been added to this network.

        Parameters
        ----------
        candidate: str
            Name of new component to be added.

        Returns
        -------
        bool
            True if the name is unique, False otherwise.

        """
        return not (candidate in self._synapse_identifiers + self._ensemble_identifiers)

    def _find_roots(self):
        """Identifies nodes that have no incoming edges. Those are used as starting points for the multi-BFS
        forward sweep, whose order also gets updated every time this method is called.

        Warning
        -------
        Roots are updated on network initialization and on every application of :meth:`add_ensemble` and
        :meth:`add_synapse`. If adding a synapse results in a rootless recurrent network, all nodes will have
        in-degree > 0. In such cases, the method will mark the first ensemble added `self.get_ensembles()[0]`
        as the root and direct external input to it. This behavior can be modified by setting `self.roots`
        directly after building the network.

        """
        updated_roots = [node for node, degree in self.graph.in_degree() if degree == 0]
        if updated_roots:
            self.roots = updated_roots

        elif self.get_ensembles():
            # if the network became rootless after adding a synapse, default to the first ensemble as the root.
            self.roots = [self.get_ensembles()[0].identifier]

        # breadth-first search (BFS) iterator used to traverse the network graph layer by layer.
        # returns a list of layers, first one consisting of the root nodes. Flatten and visit sequentially.
        self.traversal_order = flatten(list(nx.bfs_layers(self.graph, self.roots)))

    def _object_from_configuration(
        self, path: str = None, cfg: dict = None, comp_type: str = None
    ) -> (Neuron | Synapse, dict):
        """Constructs an objects from the YAML in `path` or from `cfg` dictionary if provided."""
        # load configuration dictionary from ensemble descriptor YAML.
        if path and not cfg:
            cfg = load_yaml(path)

        # initialize the ensemble/synapse based on its class reference, given in its configuration file.
        temp_import = __import__(cfg["package"], globals(), locals(), ["*"])
        class_ref = getattr(temp_import, cfg["class"])

        if comp_type == "ensemble":
            comp_ref = class_ref(**cfg["model"], configuration=cfg, device=self.device)

        else:
            comp_ref = class_ref(
                **cfg["model"],
                configuration=cfg,
                device=self.device,
                src_ensemble=self.graph.nodes[cfg["model"]["source"]]["reference"],
                dst_ensemble=self.graph.nodes[cfg["model"]["target"]]["reference"],
            )

        return comp_ref, cfg



""" Network unit and functional tests. """
import os
import pytest

from sapicore.pipeline.simple import SimpleSimulator
from sapicore.tests import ROOT

TEST_ROOT = os.path.join(ROOT, "tests", "engine", "network", "test_network")  # test root directory.


class TestNetwork:
    @pytest.mark.parametrize("cfg_", [os.path.join(TEST_ROOT, "example.yaml")], ids=["SYNTH"])
    @pytest.mark.functional
    @pytest.mark.slow
    def test_network(self, cfg_: dict):
        """All YAML-parameterized network tests go here."""
        SimpleSimulator(configuration=cfg_).run()


if __name__ == "__main__":
    pytest.main(args=["-s", "-v"])



from setuptools import setup, find_packages
from sapicore import __version__

with open("README.md") as f:
    long_description = f.read()

setup(
    name="sapicore",
    version=__version__,
    description="A framework for spiking neural network modeling. ",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cplab/sapicore",
    author="Roy Moyal, Matthew Einhorn, Ayon Borthakur, Thomas Cleland",
    author_email="rm875@cornell.edu, me263@cornell.edu, ayon.borthakur@ai.iith.ac.in, " "tac29@cornell.edu",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    packages=find_packages(),
    install_requires=[
        "h5py",
        "numpy",
        "scipy",
        "pandas",
        "torch",
        "tensorboard",
        "networkx",
        "scikit-learn",
        "matplotlib",
        "PyYAML",
        "nixio",
        "dill",
        "natsort",
        "alive_progress",
        "tree-config",
        "pytest",
        "setuptools",
    ],
    extras_require={
        "dev": ["coverage", "flake8", "sphinx<7.0.0", "sphinx-rtd-theme", "m2r2"],
    },
)



"""Neurons are units that receive, integrate, and emit signals.

Neuron instance attributes are 1D tensors of arbitrary length, initialized with a single element by default.

Neurons maintain a numeric state akin to a membrane potential in the float32 tensor `voltage`,
and sometimes a binary integer tensor `spiked` representing emitted action potential events.

All classes derived from :class:`~engine.neuron.Neuron` must implement the parent
:meth:`~engine.component.Component.forward` method.

"""
from typing import Callable

import torch
from torch import Tensor

from sapicore.engine.component import Component
from sapicore.utils.integration import Integrator, RungeKutta

__all__ = ("Neuron",)


class Neuron(Component):
    """Generic neuron base class.

    Defines and initializes instance attributes shared by all neuron objects, over and above the universal
    specifications handled by the parent class :class:`~engine.component.Component`.

    Parameters
    ----------
    equation: Callable or tuple, optional
        Anonymous function(s) specifying this neuron's difference equation(s). Provided here since all neurons
        may utilize an explicit declaration of their update rule for numeric approximation purposes.

        This is strictly optional but necessary for utilizing Sapicore's integrators. If provided, the lambda syntax
        should be used (lambda x, data: f(x, data)).

        If the Euler update rule is v[t+1] = v[t] + f(v[t])*DT, `self.equation` should be the f(v[t]) portion.
        This is because RK4 needs to compute f(v[t]) at different points, e.g. half a time step forward.

    integrator: Integrator, optional
        Specifies an :class:`~utils.integration.Integrator` to be used when approximating the next value(s)
        of this neuron's dynamic variable(s), e.g. voltage. Requires explicitly defining `equation`.
        Defaults to Runge-Kutta of order 4 to increase accuracy at a small performance penalty.

    input: Tensor
        PyTorch registered buffer tracking the input current(s) received in this simulation step.

    voltage: Tensor
        PyTorch registered buffer tracking the numeric state of the unit represented by this neuron instance,
        corresponding to a membrane voltage.

    kwargs:
        Additional instance attributes that the user may set.

    Warning
    -------
    When defining `equation` for a custom neuron model, the present value of `voltage` should NOT be added to the
    right hand side. Do NOT multiply by DT. These operations will be performed within the Integrator.

    """

    _loggable_props_: tuple[str] = ("input", "voltage")

    def __init__(self, equation: Callable = None, integrator: Integrator = None, **kwargs):
        """Initializes generic instance attributes shared by all analog and spiking neuron derived classes."""
        # register universal attributes and tracking variables shared across components.
        super().__init__(**kwargs)

        # difference equation of this neuron model, to potentially be used with numeric approximation methods.
        self.equation = equation

        # integrators, though optional, should be incorporated into forward() if used at all.
        self.integrator = RungeKutta(order=4) if integrator is None else integrator

        # register the `input` and `voltage` tensor buffers, shared by all neuron child classes.
        self.input = torch.zeros(1, dtype=torch.float, device=self.device)
        self.voltage = torch.zeros(1, dtype=torch.float, device=self.device)

        # specifies which loggable attribute should be considered this unit's output (e.g., voltage or spikes).
        self.output_field = "voltage"

        # optional gating tensor to multiply data by.
        self.gate = torch.ones_like(self.voltage)
        self.gate_ = False

        # optional teaching signal tensor to replace data by.
        self.teach = torch.ones_like(self.voltage)
        self.teach_ = False

    @property
    def num_units(self):
        """Number of functional units represented by this object.

        Neurons are singletons by coercion, as they are meant to express and encapsulate unit dynamics.
        Derivatives of :class:`~engine.ensemble.Ensemble` can modify this property and duplicate units as necessary.

        """
        return 1

    @staticmethod
    def aggregate(inputs: list[Tensor], identifiers: list[str] = None) -> Tensor:
        """Determines how presynaptic inputs from multiple sources should be aggregated.

        By default, neurons sum their inputs. However, many use cases may require more sophistication.
        Shunting inhibition, for instance, can be expressed with torch.div (or torch.prod, if the source
        synapse is expected to send the inverse).

        Parameters
        ----------
        inputs: list of Tensor
            Input arriving at this layer, synaptic or external.

        identifiers: list of str, optional
            Labels by which to micromanage input aggregation. Since some inputs may not be
            synaptic, users are responsible for passing identifiers in an order matching that of the input tensors.

        Note
        ----
        If your model requires identifier-dependent preprocessing of synaptic inputs to this neuron (e.g., a
        combination of addition and multiplication), it can be implemented by overriding this method.

        """
        return torch.sum(torch.vstack(inputs), dim=0)

    def forward(self, data: Tensor) -> dict:
        """Passes input through the neuron.

        Parameters
        ----------
        data: Tensor
            Input current whose value is used to compute the next value of the numeric state tensor `voltage`.

        Returns
        -------
        dict
            A dictionary whose keys are loggable attributes and whose values are their states as of this time step.
            For potential use by a :class:`~pipeline.simulation.SimpleSimulator` or any other
            :class:`~pipeline.Pipeline` script handling runtime operations.

        Raises
        ------
        NotImplementedError
            The forward method must be implemented by derivative classes.

        """
        raise NotImplementedError

    def inject(self, current: Tensor, mult: bool = False):
        """Injects a current into this neuron.

        Parameters
        ----------
        current: Tensor
            Float tensor containing value(s) to be added to this neuron's `voltage` tensor.

        mult: bool
            Whether the current is multiplicative. Defaults to additive (False).

        """
        if mult:
            self.voltage = self.voltage * current
        else:
            self.voltage = self.voltage + current

    def integrate(self, **kwargs) -> Tensor:
        """Generic support for numeric approximation.

        This wrapper is meant to be called from within :meth:`~engine.neuron.Neuron.forward` in the voltage
        update step. Its purpose is to take the difference equation defining a neuron and approximate the
        voltage value at time t + :attr:`~utils.constants.DT`. Keyword arguments should include variables that
        `equation` depends on. The argument `x` is the tensor's state at time t, typically `voltage`.

        See Also
        --------
        :class:`~utils.integration.RungeKutta`

        """
        return self.integrator(x=self.voltage, equation=self.equation, **kwargs)

    def gate_signal(self, signal: Tensor):
        """Inject external gating `signal`, to be multiplied by synaptic input to this neuron.

        Warning
        -------
        Has no effect unless derivative class forward method utilizes it.

        """
        self.gate = signal
        self.gate_ = True

    def teach_signal(self, signal: Tensor):
        """Inject external teaching `signal`, to replace synaptic input to this neuron.

        Warning
        -------
        Has no effect unless derivative class forward method utilizes it.

        """
        self.teach = signal
        self.teach_ = True



""" Ensembles (layers) manage neuron collections by fixing their size, extending select attribute tensors,
and handling layer-aware operations such as heterogeneous parameter initialization and normalization.

Note
----
Ensemble instance attributes are 1D tensors corresponding to potentially heterogeneous parameters
and states of neurons (tensor elements) whose behavior is prescribed by their parent class
(e.g., :class:`~engine.neuron.spiking.LIF.LIFNeuron`).

Ensembles can be initialized programmatically or from a configuration YAML containing neuron parameters and their
generation method (fixed, zipped, grid, or drawn from a distribution). Forward calls invoke the parent neuron class
method.

Example
-------
Initialize an ensemble of basic analog neurons, inheriting default parameters and behavior from
:class:`~engine.neuron.analog.AnalogNeuron`:

    >>> from sapicore.engine.ensemble.spiking import LIFEnsemble
    >>> layer = LIFEnsemble(num_units = 10)

See Also
--------
:class:`~utils.sweep.Sweep`
    For built-in parameter sweep utility methods.

"""
import torch
from torch import Tensor

from sapicore.engine.neuron import Neuron


class Ensemble(Neuron):
    """Ensemble base class. Provides generic implementations of tensor expansion and parameter diversification.

    Attributes
    ----------
    _extensible_props_: tuple[str]
        The subset of this object's configurable property names (`_config_props_`) that needs to be duplicated
        `num_units` times. Defaults to `None`, in which case all `_config_props_` are assumed extensible.
        In the rare case where no configurable properties should be extended, use an empty tuple.

    """

    _extensible_props_: tuple[str] = None

    def __init__(self, num_units: int = 1, **kwargs):
        """Constructs a generic ensemble instance, inheriting attributes from :class:`~engine.neuron.Neuron`."""
        super().__init__(**kwargs)
        self._num_units = num_units

        # expand extensible parameter tensors to size `num_units`.
        for prop in self.extensible_props:
            temp = getattr(self, prop)
            if temp is not None:
                setattr(self, prop, torch.zeros(self.num_units, dtype=torch.float, device=self.device) + temp)

        # expand loggable property tensor buffers to size `num_units`.
        for prop in self.loggable_props:
            temp = getattr(self, prop)
            if temp is not None:
                setattr(self, prop, torch.zeros(self.num_units, dtype=torch.float, device=self.device) + temp)

    @property
    def num_units(self):
        return self._num_units

    @property
    def extensible_props(self):
        if self._extensible_props_ is None:
            return self._config_props
        else:
            return self._extensible_props_

    def forward(self, data: Tensor) -> dict:
        """Processes the ensemble using the size-agnostic forward method of its neuron parent class.

        Parameters
        ----------
        data: Tensor
            External input to be processed by this ensemble.

        Returns
        -------
        dict
            Dictionary containing the numeric state tensor `voltage`, per the parent class method.

        Warning
        -------
        If population-aware operations are necessary (e.g., layer-wise normalization), those should go here.

        """
        return super().forward(data)



""" Synapse unit and integration tests. """
import os
import pytest

from sapicore.engine.synapse import Synapse
from sapicore.engine.synapse.STDP import STDPSynapse

from sapicore.tests import ROOT

STEPS = 100  # default number of simulation forward steps to execute during testing.
TEST_ROOT = os.path.join(ROOT, "tests", "synapse", "test_synapse")  # test root directory.


class TestSynapse:
    @pytest.mark.parametrize("arg_", [Synapse(), STDPSynapse()], ids=["BASE", "STDP"])
    @pytest.mark.unit
    def test_init(self, arg_: Synapse):
        """Initialization test cases."""
        # TODO write simple unit and integration tests.
        pass


if __name__ == "__main__":
    pytest.main(args=["-s", "-v"])



""" Synapses with spike-timing-dependent plasticity (STDP). """
import torch
from torch import Tensor

from sapicore.engine.synapse import Synapse

__all__ = ("STDPSynapse",)


class STDPSynapse(Synapse):
    """STDP synapse implementation.

    All STDP attributes are initialized as 2D tensors covering the source and destination ensembles, to allow for
    easily implementable heterogeneity in synaptic properties.

    STDP synapses own the following attributes, on top of those inherited from Synapse:

    Parameters
    ----------
    tau_plus: float or Tensor
        Time constant of potentiation STDP time window.

    tau_minus: float or Tensor
        Time constant of depression STDP time window.

    mu_plus: float or Tensor
        Positive dependence exponent (e.g., 1.0 for multiplicative STDP, 0.0 for additive STDP)

    mu_minus: float or Tensor
        Negative dependence exponent (e.g., 1.0 for multiplicative STDP, 0.0 for additive STDP)

    alpha_plus: float or Tensor
        Limit on magnitude of weight modification for positive spike time difference.

    alpha_minus: float or Tensor
        Limit on magnitude of weight modification for negative spike time difference.

    """

    _config_props_: tuple[str] = (
        "weight_max",
        "weight_min",
        "delay_ms",
        "tau_plus",
        "tau_minus",
        "mu_plus",
        "mu_minus",
        "alpha_plus",
        "alpha_minus",
    )
    _loggable_props_: tuple[str] = ("weights", "connections", "output")

    def __init__(
        self,
        tau_plus: float = 20.0,
        tau_minus: float = 20.0,
        mu_plus: float = 0.0,
        mu_minus: float = 0.0,
        alpha_plus: float = 0.01,
        alpha_minus: float = 0.01,
        **kwargs
    ):
        """Constructs an STDP synapse object connecting two ensembles to each other or an ensemble to itself.

        :class:`~engine.synapse.Synapse` attributes should be given as keyword arguments by the calling method
        unless default values are acceptable.
        """
        super().__init__(**kwargs)

        # configurable attributes specific to STDP, over and above a static synapse.
        self.tau_plus = torch.zeros(self.matrix_shape, dtype=torch.float, device=self.device) + tau_plus
        self.tau_minus = torch.zeros(self.matrix_shape, dtype=torch.float, device=self.device) + tau_minus

        self.mu_plus = torch.zeros(self.matrix_shape, dtype=torch.float, device=self.device) + mu_plus
        self.mu_minus = torch.zeros(self.matrix_shape, dtype=torch.float, device=self.device) + mu_minus

        self.alpha_plus = torch.zeros(self.matrix_shape, dtype=torch.float, device=self.device) + alpha_plus
        self.alpha_minus = torch.zeros(self.matrix_shape, dtype=torch.float, device=self.device) + alpha_minus

        # state attributes specific to STDP synapses, over and above Synapse.
        self.src_last_spiked = torch.zeros(self.matrix_shape[1], dtype=torch.float, device=self.device)
        self.dst_last_spiked = torch.zeros(self.matrix_shape[0], dtype=torch.float, device=self.device)

        # enable or disable learning (weight updates).
        # in typical cases, should generally be on during training and off during testing.
        self.learning = True

        # keep track of stale spike pairs, so that weights are only updated on new spike events.
        # alternatively, set `is_discontinuous` false to update weights on every simulation step.
        self.is_discontinuous = kwargs.get("is_discontinuous", True)
        self.stale = torch.ones_like(self.weights).bool()

    def update_weights(self) -> Tensor:
        """STDP weight update implementation.

        Returns
        -------
        Tensor
            2D tensor containing weight differences (dW).

        """
        # update last spike time to current simulation step iff a spike has just arrived at the synapse.
        self.src_last_spiked = self.src_last_spiked + self.delayed_data * (-self.src_last_spiked + self.simulation_step)

        # update last spike time only for postsynaptic units that spiked in this model iteration.
        self.dst_last_spiked = self.dst_last_spiked + self.dst_ensemble.spiked * (
            -self.dst_last_spiked + self.simulation_step
        )

        # ensure action is only taken on pre-post pairs for whom one or both units spiked in this cycle.
        if self.is_discontinuous:
            self.stale[:] = True
            self.stale[:, torch.argwhere(torch.as_tensor(self.src_last_spiked == self.simulation_step))] = False
            self.stale[torch.argwhere(torch.as_tensor(self.dst_last_spiked == self.simulation_step)), :] = False
        else:
            self.stale[:] = False

        # initialize delta weight matrix (destination X source).
        delta_weight = torch.zeros(self.matrix_shape, dtype=torch.float, device=self.device)

        if torch.any(~self.stale) or not self.is_discontinuous:
            # subtract postsynaptic last spike time stamps from presynaptic.
            # transpose dst_last_spiked to a column vector, extend column-wise, then subtract from src_last_spiked.
            dst_tr = self.dst_last_spiked.reshape(self.dst_last_spiked.shape[0], 1)
            delta_spike = -dst_tr.repeat(dst_tr.shape[1], self.src_ensemble.num_units) + self.src_last_spiked

            # spike time differences for the potentiation and depression cases (pre < post, pre > post, respectively).
            ltp_diffs = (delta_spike < 0.0).int() * delta_spike
            ltd_diffs = (delta_spike > 0.0).int() * delta_spike

            # add to total delta weight matrix (diffs are in simulation steps, tau are in ms).
            delta_weight = delta_weight + ~self.stale * (delta_spike < 0.0).int() * (
                self.alpha_plus * torch.exp(ltp_diffs / (self.tau_plus / self.dt))
            )
            delta_weight = delta_weight + ~self.stale * (delta_spike > 0.0).int() * (
                -self.alpha_minus * torch.exp(-ltd_diffs / (self.tau_minus / self.dt))
            )
            self.weights = self.weights.add(delta_weight)

        return delta_weight

    def forward(self, data: Tensor) -> dict:
        """Updates weights if need be, then calls the parent :class:`~engine.synapse.Synapse` :meth:`forward` method.

        Parameters
        ----------
        data: Tensor
            Source ensemble spike output to be processed by this STDP synapse.

        Returns
        -------
        dict
            Dictionary containing `weights`, `connections`, and `output` for further processing.

        """
        if self.learning:
            self.update_weights()

        return super().forward(data)
